/*
      *****************************************************************************
      *  Copyright (C) 2017 by Reinhard Heuberger                                 *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
*/

// Sector:    --____---------------------------------------------------____--
// Format:          |<Servo/Header>|<------Data -----------Data------>|
// Write-Gate _____________________|----------------------------------|______
//                                 |<<< DEC_clk, MFM-DEcoder clock >>>|
//                  |<<<<<<<<<<<<<<<<<< ENC_clk, MFM-ENcoder clock >>>|
//
//  by: Reinhard Heuberger , www.PDP11GY.com, AUG, 2011
//  Improved Version for the DE10-Lite board
//
//
module my_clock_domain_mux_DE10 (
       clk_65_6,                   // FPGA System  Clock  65.6 MHz
       clk_ENcoder,                // 16Bit clock @ MFM ENcoder
       clk_DEcoder,                // 16Bit clock @ MFM DEcoder
       Write_gate,                 // Write Gate
       //
       DPR_clk,                    // Dual_Ported-RAM Clock, switched=glitchfree
       SEQ_clk);                   // Sequencer Clock, switched=glitchfree
//
input clk_65_6;
input clk_ENcoder;
input clk_DEcoder;
input Write_gate;
output DPR_clk, SEQ_clk;      
//
reg    DPR_clk, SEQ_clk, SO;
reg [7:0]  DEclk_shifter;
reg [7:0]  ENclk_shifter;
wire ENclk, DEclk, COclk; 
//
initial
begin
   DEclk_shifter <= 0;
   ENclk_shifter <= 0;
end
//
always @ (posedge clk_65_6)
 begin
  DEclk_shifter = {DEclk_shifter[6:0], ~clk_DEcoder};   // shift
 end
//
always @ (posedge clk_65_6)
begin
 ENclk_shifter = {ENclk_shifter[6:0], ~clk_ENcoder};   // shift
end
//
assign DEclk =  DEclk_shifter[4];   // "ok"
assign ENclk = ~DEclk_shifter[0];   // "ok"
//
always @ (posedge clk_65_6)
 begin
   if (Write_gate) begin
     DPR_clk <= ENclk;
     SEQ_clk <= DEclk;
   end else if (!Write_gate) begin
     DPR_clk <= clk_ENcoder;   // "ok"
     SEQ_clk <= clk_ENcoder;   // "ok"
   end
 end
endmodule
