/*
      *****************************************************************************
      *  Copyright (C) 2017 by Reinhard Heuberger                                 *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
*/
//
//
//  by: Reinhard Heuberger , www.PDP11GY.com, AUG, 2011
//  MAI 2017 : Addons for the MAX10/DE10-Lite board 
//
module my_drive_select_DE10 (
       enable,              // Enable
       clk_65_6,            // Clock
       Drive0_set,          ///
       Drive1_set,          ////
		 Drive2_set,          /////
		 Drive3_set,          //////-> Dip Switch
       Drive0_in,           //
       Drive1_in,           //
		 secpuls,             //
                            //
		 match0_H,
		 match1_H,
		 match2_H,
		 match3_H,
       match_H,      		 //
		 dl3_segment,
		 dl2_segment,
		 dl1_segment,
		 dl0_segment,
		 heartbeat);          //
//
input  enable, clk_65_6;
input  Drive0_set, Drive1_set, Drive2_set, Drive3_set;
input  Drive0_in, Drive1_in, secpuls;
output match0_H, match1_H, match2_H, match3_H;
output match_H, heartbeat;
output [7:0] dl3_segment;
output [7:0] dl2_segment;
output [7:0] dl1_segment;
output [7:0] dl0_segment;
//
reg match_H, match0_H, match1_H, match2_H, match3_H, secS, seg_7;
reg [31:0]  divider;
reg [8:0]   shifter;
reg [7:0]   dl0_segment;
reg [7:0]   dl1_segment;
reg [7:0]   dl2_segment;
reg [7:0]   dl3_segment;
//
initial
begin
	match0_H <= 0;
   match1_H <= 0;
   match2_H <= 0;
   match3_H <= 0;
   match_H  <= 0;
	shifter   = 1;
end
//
always @ (posedge clk_65_6)
begin
  if(enable)begin
    if(Drive0_set && Drive1_in && Drive0_in)         // drive 0 
	   begin
	    match0_H <= 1;
		 match1_H <= 0;
		 match2_H <= 0;
		 match3_H <= 0;
		 match_H  <= 1;
		end
	 else if(Drive1_set && Drive1_in && !Drive0_in)   // drive 1
	   begin
	    match0_H <= 0;
		 match1_H <= 1;
		 match2_H <= 0;
		 match3_H <= 0;
		 match_H  <= 1;
		end
	 else if(Drive2_set && !Drive1_in && Drive0_in)    // drive 2
	   begin
	    match0_H <= 0;
		 match1_H <= 0;
		 match2_H <= 1;
		 match3_H <= 0;
		 match_H  <= 1;		
		end
	 else if(Drive3_set && !Drive1_in && !Drive0_in)   // drive 3
	   begin
	    match0_H <= 0;
		 match1_H <= 0;
		 match2_H <= 0;
		 match3_H <= 1;
		 match_H  <= 1;
		end
	 else 
	    match_H <= 0;
  end else begin
    match_H <= 0;
  end
end
//
//
always @ (posedge clk_65_6)
begin
 secS <= secpuls;
end
//
always @ (posedge clk_65_6)
begin
  if ( secpuls  & !secS) begin
    divider <= divider +1 ;
  end else begin
   divider <= divider ;
  end
end
assign heartbeat = divider[9];
//
always @ (posedge clk_65_6)
begin
 seg_7 <= divider[4];
end
//
always @ (posedge clk_65_6)
begin
  if ( divider[4]  & !seg_7) begin
   if (Drive0_set ) begin
     dl0_segment[5:0] <= ~shifter[6:1];    // dl0 set, drive 0
	  dl0_segment[6] <= 1;
	 end else begin
	  dl0_segment[6]    <= 0;
	  dl0_segment[5:0]  <= 63;
	 end
	 //
   if (Drive1_set ) begin
     dl1_segment[5:0] <= ~shifter[6:1];    // dl1 set, drive 1
	  dl1_segment[6] <= 1;
	 end else begin
	  dl1_segment[6]    <= 0;
	  dl1_segment[5:0]  <= 63;
	 end
	 //
   if (Drive2_set ) begin
     dl2_segment[5:0] <= ~shifter[6:1];    // dl2 set, drive 2
	  dl2_segment[6] <= 1;
	 end else begin
	  dl2_segment[6]    <= 0;
	  dl2_segment[5:0]  <= 63;
	 end
	 //
   if (Drive3_set ) begin
     dl3_segment[5:0] <= ~shifter[6:1];    // dl3 set, drive 3
	  dl3_segment[6] <= 1;
	 end else begin
	  dl3_segment[6]    <= 0;
	  dl3_segment[5:0]  <= 63;
	 end
	 //	 
   shifter = shifter << 1;              // shift
   shifter[0] <= shifter[6];            // rotate
  end
end
//
endmodule
